import os
import json
import shutil
import platform
import pandas as pd
import tomli
from datetime import datetime, timedelta
from pathlib import Path


def initCowBase_farm(rootdir):
    """
    Initializes the folder structure in which the CowBase package will be mainly operating.

    Parameters
    ----------
    rootdir : str
        Root directory in which the CowBase folder structure should be implemented in.

    Notes
    ----------
    A number of folders and files is created when initializing:
    1. CowBase - parent directory
    2. CowBase/INPUT - folder in which the AMS (automatic milking system) backup files (.bak.zip/.zip) should be saved
    3. CowBase/INPUT/d_farmname - Backups from DeLaval systems should always be saved in folders starting with 'd_' and followed by the farmname
    4. CowBase/INPUT/l_farmname - Backups from Lely systems should always be saved in folders starting with 'l_' and followed by the farmname
    5. CowBase/OUTPUT - folder in which the unpacked data files (.csv) will be later saved
    6. CowBase/temp - folder in which temparary files are stored
    7. CowBase/config - folder in which a user config file and dictionaries are stored

    """

    rootdir = Path(rootdir)

    # initialize folder structure
    rootdir_cowbase = rootdir / "CowBase"
    rootdir_cowbase.mkdir(exist_ok=True, parents=True)

    rootdir_OUTPUT = rootdir_cowbase / "OUTPUT"
    rootdir_OUTPUT.mkdir(exist_ok=True)

    rootdir_config = rootdir_cowbase / "config"
    rootdir_config.mkdir(exist_ok=True)

    # add config and tableDict to the config folder for individual settings
    for config_file in (Path(__file__).parent / "utils").glob("*"):
        # If the config file is already in dir, don't copy again
        if (rootdir_config / config_file.name).exists():
            continue

        shutil.copy(
            config_file,
            rootdir_config / config_file.name,
        )

    # set the rootdir in the config file to the defined rootdir
    # TODO: if the config file has already been edited, don't overwrite
    osname = os.name
    fin = open((Path(__file__).parent / "utils" / "config.toml"), "rt")
    fout = open(f"{rootdir_config / 'config.toml'}", "wt")
    for line in fin:
        # read replace the string and write to output file
        if osname == "nt":
            hostname = platform.node()
            fout.write(
                line.replace("SERVERNAME", f"{hostname}\SQLEXPRESS").replace(
                    "rootdir_cowbase", str(rootdir_cowbase)
                )
            )
        elif osname == "posix":
            fout.write(
                line.replace("SERVERNAME", "localhost").replace(
                    "rootdir_cowbase", str(rootdir_cowbase)
                )
            )
    # close input and output files
    fin.close()
    fout.close()
    dtransfer_summary = pd.DataFrame(
        {
            "timestamp": ["20230101000000"],
            "OID": [1],
            "OID2": [1],
            "issaved": [1],
            "issend": [1],
            "isdel": [1],
        }
    )
    dtransfer_summary.to_csv(rootdir_config / "cb_data_summary.csv", index=False)


def farmExtract(
    filepath_config="/var/lib/CowBase/config/config_farm.toml",
    filepath_tdict="/var/lib/CowBase/config/M2_tableDict.json",
):
    """
    Extracts data tables from restored MSQL database

    Parameters
    ----------
    tabledict : dict
        dictonary that contains all table and column names from the respective milking system that should be extracted
    table : str
        tablename of the table that should be restored
    robottype : str
        brand of milking system the backup is from (d-DeLaval; l-Lely)
    dbname : str
        name of the database
    filename_out : str
        filepath and name to which the extracted data will be saved to
    config : dict
        config file with user settings

    """
    with open(filepath_config, mode="rb") as ctomli:
        config = tomli.load(ctomli)
    with open(filepath_tdict) as file:
        tabledict = json.load(file)

    dtransfer_summary = pd.read_csv(
        config["user"]["rootdir"] / "config" / "cb_data_summary.csv"
    )
    lastoid = dtransfer_summary.OID.max()
    timenow = datetime.now().strftime("%Y%m%d%H%M")
    timebackups = (datetime.now() + timedelta(days=-1)).strftime("(%Y-%m-%d)")

    # initialize folder structure
    rootdir_OUTPUT = config["user"]["rootdir"] / "OUTPUT"
    rootdir_OUTPUT2 = rootdir_OUTPUT / "01_full_table"
    rootdir_OUTPUT2.mkdir(exist_ok=True)
    rootdir_filedest = rootdir_OUTPUT2 / timenow
    rootdir_filedest.mkdir(exist_ok=True)

    for table in tabledict[config["user"]["robottype"]]:
        if tabledict[config["user"]["robottype"]][table]["extracttype"][0] == "full":
            filename_out = (
                rootdir_filedest
                / f'{config["user"]["robottype"]}_{config["user"]["farmname"]}_{timenow}_{table}.csv'
            )
            columnnames = ", ".join(
                tabledict[config["user"]["robottype"]][table]["columns"]
            )
            # export table
            sql = f'SELECT {columnnames} FROM {config["user"]["dbname"]}.dbo.{table}'
        elif (
            tabledict[config["user"]["robottype"]][table]["extracttype"][0]
            == "datetime"
        ):
            filename_out = (
                rootdir_filedest
                / f'{config["user"]["robottype"]}_{config["user"]["farmname"]}_{timenow}_{table}.csv'
            )
            columnnames = ", ".join(
                tabledict[config["user"]["robottype"]][table]["columns"]
            )
            # export table
            sql = f'SELECT {columnnames} FROM {config["user"]["dbname"]}.dbo.{table} WHERE {tabledict[config["user"]["robottype"]][table]["restrictID"][0]} > {timebackups}'
        elif tabledict[config["user"]["robottype"]][table]["extracttype"][0] == "moid":
            filename_out = (
                rootdir_filedest
                / f'{config["user"]["robottype"]}_{config["user"]["farmname"]}_{timenow}_{table}.csv'
            )
            columnnames = ", ".join(
                tabledict[config["user"]["robottype"]][table]["columns"]
            )
            # export table
            sql = f'SELECT {columnnames} FROM {config["user"]["dbname"]}.dbo.{table} WHERE OID > {lastoid}'
        cmd = (
            'bcp "'
            + sql
            + f'" queryout {filename_out} -d{config["user"]["dbname"]} -c -C 65001 -t"^^" -m 9999  -b 10000 -a 10000 -k -T'
        )
        os.system(cmd)

        try:
            datatable = pd.read_csv(
                filename_out, header=None, delimiter="\^\^", engine="python"
            )
            datatable = datatable.replace({";": ","}, regex=True)
            datatable.to_csv(
                filename_out,
                header=list(tabledict[config["user"]["robottype"]][table]["columns"]),
                index=False,
                sep=";",
                escapechar="\\",
            )
        except pd.errors.EmptyDataError:
            print(f"Skipped! No data available for {table}!")

        if table == "VoluntarySessionMilkYield":
            # Select largest OID from VoluntarySessionMilkYield table
            try:
                df_vsmy = pd.read_csv(filename_out, delimiter=";")
                newoid = df_vsmy.OID.max()
            except pd.errors.EmptyDataError:
                newoid = lastoid

    dtransfer_newentry = pd.DataFrame(
        {
            "timestamp": [timenow],
            "OID": [lastoid],
            "OID2": [newoid],
            "issaved": [1],
            "issend": [0],
            "isdel": [0],
        }
    )

    dtransfer_summary = pd.concat(
        [dtransfer_summary, dtransfer_newentry], ignore_index=True
    )

    dtransfer_summary.to_csv(
        config["user"]["rootdir"] / "config" / "cb_data_summary.csv",
        index=False,
    )


# TODO: add a function for file transfer
def cb_filetransfer():
    """
    Function to handle the data transfer between the farm PC and a research server where the data should be send to
    (ftp, https, ...?)
    """



import datetime as dt
import json
import os
import random
import re

import numpy as np
import pandas as pd

from cowbase.cowbase_m5 import DB_connect

import sqlite3
from pathlib import Path
from meteostat import Point, Hourly, Stations


def initiateDatabase(rootdir: str | Path, dbtype: str) -> None:
    """
    Creates a new database using CowBase's data scheme

    Parameters
    ----------
    rootdir : str or Path
        Filepath to CowBase root directory (e.g. linux: '/var/lib/'; windows: 'C:\\Users\\user\\Documents\\')
    dbtype : str
        postgres, mysql, sqlite - database management software needs to be downloaded and installed prior to this
    """

    rootdir = Path(rootdir)

    with open(rootdir / "config" / "serverSettings.json") as file:
        serverSettings = json.load(file)
    with open(rootdir / "config" / "M4_tableDict.json") as file:
        tableDict = json.load(file)

    db_connect = DB_connect(**serverSettings)
    db_connect.create_db()

    dbtype = serverSettings["dbtype"]

    for tables in tableDict:
        sql_statement = f"CREATE TABLE IF NOT EXISTS {tables} ("
        for columnnames in tableDict[tables][dbtype]:
            sql_statement += f"{columnnames} {tableDict[tables][dbtype][columnnames]}, "
        for pkeys in tableDict[tables]["p_key"]:
            sql_statement += f"PRIMARY KEY ({pkeys}), "
        for fkeys in tableDict[tables]["f_keys"]:
            sql_statement += f"FOREIGN KEY ({fkeys}) REFERENCES {tableDict[tables]['f_keys'][fkeys]}, "
        sql_statement += f"UNIQUE({', '.join(tableDict[tables]['unique'])}), "
        sql_statement = sql_statement[0:-2] + ");"
        db_connect.execute(query=sql_statement)

    if dbtype == "postgres":
        sql_statement = """
        do
        $$
        begin
            if not exists (select * from pg_user where usename = 'dbview') then
                create role dbview password 'view';
            end if;
        end
        $$
        ;
        do
        $$
        begin
            if not exists (select * from pg_user where usename = 'dbwrite') then
                create role dbwrite password 'write';
            end if;
        end
        $$
        ;
        GRANT SELECT ON ALL TABLES IN SCHEMA public TO dbview;
        GRANT SELECT, INSERT, UPDATE ON ALL TABLES IN SCHEMA public TO dbwrite;
        """
        db_connect.execute(query=sql_statement)


def create_sql_insert_update(db_update: dict) -> str:
    """
    Creates an upsert sql statement from a dictionary containing necessary parameters (table, insert, conflict, update)

    Parameters
    ----------
    db_update : dict
        dictionary containing information to generate a sql statement

    Return
    ------
    update_sql : str
        sql statement
    """
    update_sql = f'INSERT INTO {db_update["table"]}'

    if db_update["insert"]:
        update_sql += f" ("
        for elem in db_update["insert"]:
            update_sql += f"{elem}, "
        update_sql = update_sql[:-2] + f") VALUES ("
        for elem in db_update["insert"]:
            update_sql += f":{elem}, "
        update_sql = update_sql[:-2] + f") "

    if db_update["conflict"]:
        update_sql += f"ON CONFLICT ("
        for elem in db_update["conflict"]:
            update_sql += f"{elem}, "
        update_sql = update_sql[:-2] + f") "

    if db_update["update"]:
        update_sql += f"DO UPDATE SET "
        for elem in db_update["update"]:
            update_sql += f"{elem} = EXCLUDED.{elem}, "
        update_sql = update_sql[:-2]

    update_sql += ";"
    return update_sql


def create_sql_update(db_update: dict) -> str:
    """
    Creates an update sql statement from a dictionary containing necessary parameters (table, update, where)

    Parameters
    ----------
    db_update : dict
        dictionary containing information to generate a sql statement

    Return
    ------
    update_sql : str
        sql statement
    """
    update_sql = f'UPDATE {db_update["table"]}'

    if db_update["update"]:
        update_sql += f" SET "
        for elem in db_update["update"]:
            update_sql += f"{elem} = :{elem}, "
        update_sql = update_sql[:-2]

    if db_update["where"]:
        update_sql += f" WHERE "
        for elem in db_update["where"]:
            update_sql += f"{elem} = :{elem} AND "
        update_sql = update_sql[:-4]

    update_sql += ";"
    return update_sql


def readData(
    filepath: str | Path, tablename: str, tableDict: dict, farmtype: str
) -> pd.DataFrame:
    """
    Read data from file (pandas)

    Parameters
    ----------
    filepath : str
        path to datafile (.csv)
    tablename : str
        name of the table/file
    tableDict : dict
        dictionary containing information about the table schema of the target database
    farmtype : str
        milking system identifier (d- delaval, l- lely)

    Return
    ------
    df_data : pandas dataframe

    """
    df_data = pd.read_csv(filepath, delimiter=";", header=0, engine="python")

    notnull_columns = tableDict[tablename][f"not_null_read"]

    if len(notnull_columns) > 0:
        for column in list(notnull_columns):
            df_data = df_data.loc[~pd.isnull(df_data[column])]

    datetime_columns = tableDict[tablename][f"{farmtype}_datetime"]

    if len(datetime_columns) > 0:
        for datetime_column in datetime_columns:
            df_data[datetime_column] = df_data[datetime_column].apply(pd.to_datetime)
            df_data[datetime_column] = df_data[datetime_column].dt.floor("s")
    return df_data


def updateFarm(
    df_data: pd.DataFrame,
    db_connect: DB_connect,
    farmname: str,
    milkingSystem: str,
    sqlupdateDict: dict,
) -> tuple[pd.DataFrame, int]:
    """
    Update farm information

    Parameters
    ----------
    df_data : pandas dataframe
        dataframe for a single table
    db_connect : sql engine
        engine to connect ot database
    farmname : str
        name of farm for which data is processed
    milkingSystem : str
        milking system identifier (d- delaval, l- lely)
    sqlupdateDict : dict
        dictionary containing information to generate a sql statement

    Return
    ------
    df_data : pandas dataframe
    farm_id : int

    """

    farm_sql = """
    SELECT farm_id, farmname, created_on
    FROM farm
    ;
    """
    farm_db = db_connect.query(query=farm_sql)

    if farmname in farm_db.loc[:, "farmname"].tolist():
        farm_update = pd.DataFrame(
            {
                "farmname": [str(farmname)],
                "created_on": [
                    farm_db.loc[farm_db["farmname"] == farmname, "created_on"].iloc[0]
                ],
                "updated_on": [dt.date.today()],
                "milking_system_type": [milkingSystem],
            }
        )

        # Update database table 'farm' with data from .txt file - insert if no data entry exist yet, update else

        farm_update_sql = create_sql_update(sqlupdateDict["sql_update"]["farm_upd"])
        db_connect.insert(query=farm_update_sql, data=farm_update.to_dict("records"))

    else:
        farm_update = pd.DataFrame(
            {
                "farmname": [str(farmname)],
                "created_on": [dt.date.today()],
                "updated_on": [dt.date.today()],
                "milking_system_type": [milkingSystem],
            }
        )

        # Insert into database table 'farm' with data from .txt file - insert if no data entry exist yet, update else

        farm_update_sql = create_sql_insert_update(sqlupdateDict["sql_update"]["farm"])
        db_connect.insert(query=farm_update_sql, data=farm_update.to_dict("records"))

    farm_id = db_connect.query(
        query=f"""SELECT farm_id FROM farm WHERE farmname = '{farmname}';"""
    ).squeeze()

    df_data["farm_id"] = farm_id
    return df_data, farm_id


def updateAnimal(
    df_data: pd.DataFrame, db_connect: DB_connect, farm_id: int, sqlupdateDict: dict
) -> pd.DataFrame:
    """
    Update animal table

    Parameters
    ----------
    df_data : pandas dataframe
        dataframe for a single table
    db_connect : sql engine
        engine to connect ot database
    farm_id : integer
        id of farm for which data is processed
    sqlupdateDict : dict
        dictionary containing information to generate a sql statement
    """

    # animal update
    animal_sql = """
    SELECT farm_id, animal_oid
    FROM animal
    ;
    """
    animal_db = db_connect.query(query=animal_sql)

    df_check_animal = df_data.copy(deep=True)
    df_check_animal = df_check_animal[["farm_id", "animal_oid"]].drop_duplicates()

    df_check_animal = df_check_animal.loc[~pd.isnull(df_check_animal.animal_oid)]
    df_check_animal.animal_oid = df_check_animal.animal_oid.astype(int)

    df_animal_update = df_check_animal.merge(
        animal_db[["farm_id", "animal_oid"]],
        how="left",
        on=["farm_id", "animal_oid"],
        indicator=True,
    )

    df_animal_update = df_animal_update.loc[
        df_animal_update["_merge"] == "left_only"
    ].drop(columns=["_merge"])

    df_animal_update = df_animal_update.loc[~pd.isnull(df_animal_update.animal_oid)]

    df_animal_update.animal_oid = df_animal_update.animal_oid.astype(int)

    df_animal_update["updated_on"] = dt.date.today()

    if len(df_animal_update) >= 1:
        # Update database table 'animal' with data from lactation data - insert if no data entry exist yet, update else

        df_animal_update = df_animal_update.drop(columns=["animal_id"])
        animal_update_sql = create_sql_insert_update(
            sqlupdateDict["sql_update"]["animal_table"]
        )
        db_connect.insert(
            query=animal_update_sql, data=df_animal_update.to_dict("records")
        )

    animal_ids = db_connect.query(
        query=f"""SELECT animal_id, animal_oid FROM animal WHERE farm_id = {farm_id};"""
    )

    dict_animal_ids = pd.Series(
        animal_ids.animal_id.values, index=animal_ids.animal_oid
    ).to_dict()

    df_data["animal_id"] = df_data["animal_oid"].map(dict_animal_ids)

    df_data = df_data.drop(columns=["animal_oid"])

    return df_data


def checkDatabase(
    df_data: pd.DataFrame,
    db_connect: DB_connect,
    table: str,
    farm_id: int,
    tableDict: dict,
) -> pd.DataFrame:
    """
    Check database for already present data

    Parameters
    ----------
    df_data : pandas dataframe
        dataframe for a single table
    db_connect : sql engine
        engine to connect ot database
    table : str
        name of the table/file
    farm_id : integer
        id of farm for which data is processed
    tableDict : dict
        dictionary containing information about the table schema of the target database

    Return
    ------
    data_update_new : pandas dataframe

    """
    data_sql = f"""
    SELECT {', '.join(tableDict[table]['unique'])}
    FROM {table}
    WHERE farm_id = {farm_id}
    ;
    """
    data_db = db_connect.query(query=data_sql)

    df_data = df_data.drop_duplicates(subset=tableDict[table]["unique"])

    data_update_new = df_data.merge(
        data_db,
        how="left",
        on=tableDict[table]["unique"],
        indicator=True,
    )

    data_update_new = data_update_new.loc[
        data_update_new["_merge"] == "left_only"
    ].drop(columns=["_merge"])

    return data_update_new


def writeData(
    df_data: pd.DataFrame, db_connect, table: str, tableDict: dict, farmtype: str
) -> None:
    """
    Write new data to 'table'

    Parameters
    ----------
    df_data : pandas dataframe
        dataframe for a single table
    db_connect : sql engine
        engine to connect ot database
    table : str
        name of the table/file
    tableDict : dict
        dictionary containing information about the table schema of the target database
    farmtype : str
        milking system identifier (d- delaval, l- lely)
    """
    df_data["updated_on"] = dt.date.today()

    datetime_columns = tableDict[table][f"{farmtype}_datetime"]
    if len(datetime_columns) > 0:
        for column in datetime_columns:
            df_data[column] = df_data[column].dt.strftime("%Y-%m-%d %H:%M:%S")

    notnull_columns = tableDict[table][f"not_null_write"]

    if len(notnull_columns) > 0:
        for column in list(notnull_columns):
            df_data_remove = df_data.loc[pd.isnull(df_data[column])]
            df_data = df_data.loc[~pd.isnull(df_data[column])]
            if len(df_data_remove) > 0:
                print('NOT NULL constraint failed; Data has been removed!')
                print(df_data_remove)

    df_data.to_sql(
        f"{table}", con=db_connect.ret_con(), if_exists="append", index=False
    )


def updateMilking(
    df_milking: pd.DataFrame,
    db_connect,
    farm_id: int,
    farmname: str,
    farmtype: str,
    sqlupdateDict: dict,
    tableDict: dict,
) -> None:
    """
    Update milking table

    Parameters
    ----------
    df_milking : pandas dataframe
        dataframe for the milking table
    db_connect : sql engine
        engine to connect ot database
    farm_id : integer
        id of farm for which data is processed
    farmname : str
        name of farm for which data is processed
    farmtype : str
        milking system identifier (d- delaval, l- lely)
    sqlupdateDict : dict
        dictionary containing information to generate a sql statement
    tableDict : dict
        dictionary containing information about the table schema of the target database
    """
    milking_sql = f"""
    SELECT farm_id, animal_id, milking_oid, started_at, ended_at
    FROM milking
    WHERE farm_id = {farm_id}
    ;
    """
    milking_db = db_connect.query(query=milking_sql)

    df_milking["farm_id"] = farm_id

    df_milking = df_milking.drop_duplicates(
        subset=["farm_id", "animal_id", "milking_oid", "started_at", "ended_at"]
    )

    milking_update_new = df_milking.merge(
        milking_db[["farm_id", "animal_id", "milking_oid"]],
        how="left",
        on=["farm_id", "animal_id", "milking_oid"],
        indicator=True,
    )

    milking_update_new.started_at = milking_update_new.started_at.apply(pd.to_datetime)
    milking_update_new.ended_at = milking_update_new.ended_at.apply(pd.to_datetime)

    ################################################################################################################
    # apply lactation_ids, parities and calving dates to milking_update_new

    lactation_sql = f"""
    SELECT farm_id, animal_id, lactation_oid, calving, parity
    FROM lactation
    WHERE farm_id = {farm_id}
    ;
    """
    lactation_db = db_connect.query(query=lactation_sql)

    if len(milking_update_new) > 0:
        for lacids in sorted(lactation_db.lactation_oid.unique()):
            aniid = lactation_db.loc[
                lactation_db.lactation_oid == lacids, "animal_id"
            ].values[0]
            calvingdate = (
                lactation_db.loc[lactation_db.lactation_oid == lacids, "calving"]
                .apply(pd.to_datetime)
                .values[0]
            )
            if calvingdate == None:  # fixing error when calvingdate == None
                calvingdate = milking_update_new.loc[
                    (milking_update_new.animal_id == aniid), "started_at"
                ].min()

            lacno = lactation_db.loc[
                lactation_db.lactation_oid == lacids, "parity"
            ].values[0]
            milking_update_new.loc[
                (milking_update_new.animal_id == aniid)
                & (milking_update_new.started_at >= calvingdate),
                "lactation_oid_db",
            ] = lacids
            milking_update_new.loc[
                (milking_update_new.animal_id == aniid)
                & (milking_update_new.started_at >= calvingdate),
                "parity_db",
            ] = lacno
            milking_update_new.loc[
                (milking_update_new.animal_id == aniid)
                & (milking_update_new.started_at >= calvingdate),
                "calving_db",
            ] = calvingdate

        ################################################################################################################
        # correct for missed/wrong lactation ids

        # delete if no milking_oid available
        milking_update_new = milking_update_new.loc[
            ~pd.isnull(milking_update_new["milking_oid"])
        ]

        # sort milk data and calculate gaps
        milking_update_new = milking_update_new.sort_values(
            by=["animal_id", "started_at"]
        ).reset_index(drop=True)
        milking_update_new["gap"] = (
            milking_update_new.groupby(by="animal_id")
            .started_at.diff()
            .dt.total_seconds()
            / 86400
        )

        milking_update_new.loc[pd.isnull(milking_update_new.gap), "gap"] = 0
        milking_update_new.loc[milking_update_new.gap < 10, "gap"] = 0
        milking_update_new.loc[milking_update_new.gap > 0, "gap"] = 1
        if farmtype == "d":
            if farmname in list(sqlupdateDict["farminfo"]["lacgapstart"]):
                milking_update_new.loc[
                    (
                        milking_update_new.started_at
                        < dt.datetime.strptime(
                            sqlupdateDict["farminfo"]["lacgapstart"][farmname],
                            "%d-%m-%Y",
                        )
                    )
                    & (
                        (
                            milking_update_new.started_at
                            - milking_update_new.calving_db
                        ).dt.total_seconds()
                        / 86400
                        < 300
                    ),
                    "gap",
                ] = 0
            else:
                startingdate = df_milking.started_at.max() - dt.timedelta(180)
                milking_update_new.loc[
                    (milking_update_new.started_at < startingdate)
                    & (
                        (
                            milking_update_new.started_at
                            - milking_update_new.calving_db
                        ).dt.total_seconds()
                        / 86400
                        < 300
                    ),
                    "gap",
                ] = 0
        # define periods restricted by gaps of at least 10 days
        milking_update_new["periods"] = milking_update_new.groupby(
            by="animal_id"
        ).gap.cumsum()
        # introduce parameter that gives 1 to periods that were fully in a period that we trust (started after data was contiouusly collected)
        milking_update_new.loc[milking_update_new.periods == 0, "trust_cal"] = 0
        milking_update_new.loc[milking_update_new.periods > 0, "trust_cal"] = 1
        # define the calculated calving date as the date of the first milking in each period
        milking_update_new["calving_cal"] = milking_update_new.groupby(
            by=["animal_id", "periods"]
        ).started_at.transform("min")
        milking_update_new["calving_cal"] = milking_update_new["calving_cal"].dt.date
        # set the parity to 1 if the parity was 0
        milking_update_new.loc[
            (milking_update_new.parity_db == 0)
            | pd.isnull(milking_update_new.parity_db),
            "parity_db",
        ] = 1
        # the calculated parity is the sum of the minimal parity for each cow and the period counter calculated based on gaps of min 10 days
        milking_update_new["parity"] = milking_update_new.groupby(
            by=["animal_id"]
        ).parity_db.transform("min")
        milking_update_new["parity"] = (
            milking_update_new["parity"] + milking_update_new["periods"]
        )
        # if there were gaps of less than 10 days between calvings (and those were recorded), correct for that
        for wp_ani_id in milking_update_new.loc[
            milking_update_new.parity_db > milking_update_new.parity, "animal_id"
        ].unique():
            while (
                milking_update_new.loc[
                    (milking_update_new.animal_id == wp_ani_id)
                    & (milking_update_new.parity_db > milking_update_new.parity)
                ].empty
                == False
            ):
                inc_time = milking_update_new.loc[
                    (milking_update_new.animal_id == wp_ani_id)
                    & (milking_update_new.parity_db > milking_update_new.parity),
                    "started_at",
                ].min()
                milking_update_new.loc[
                    (milking_update_new.animal_id == wp_ani_id)
                    & (milking_update_new.started_at >= inc_time),
                    "parity",
                ] += 1

        # the final calvingdate and lactation_oid is defined as:
        # taken from the farm database if the calculated and the predicted parity are the same and the calvingdate is before the first milking of that period
        # as the first day in the milking otherwise

        # if the calving was before the data where continouus data was collected, or
        # if the parity is the same for calculated and in the database and the calving in the database is before the calculated, trust the database
        milking_update_new.loc[
            (milking_update_new.parity_db == milking_update_new.parity)
            & (milking_update_new.calving_db <= milking_update_new.calving_cal)
            | (milking_update_new.trust_cal == 0),
            "calving",
        ] = milking_update_new.loc[
            (milking_update_new.parity_db == milking_update_new.parity)
            & (milking_update_new.calving_db <= milking_update_new.calving_cal)
            | (milking_update_new.trust_cal == 0),
            "calving_db",
        ].apply(
            pd.to_datetime
        )
        milking_update_new.loc[
            (milking_update_new.parity_db == milking_update_new.parity)
            & (milking_update_new.calving_db <= milking_update_new.calving_cal)
            | (milking_update_new.trust_cal == 0),
            "lactation_oid",
        ] = milking_update_new.loc[
            (milking_update_new.parity_db == milking_update_new.parity)
            & (milking_update_new.calving_db <= milking_update_new.calving_cal)
            | (milking_update_new.trust_cal == 0),
            "lactation_oid_db",
        ]

        # Otherwise, trust the calculated values
        milking_update_new.loc[
            (milking_update_new.parity_db == milking_update_new.parity)
            & (milking_update_new.calving_db > milking_update_new.calving_cal)
            & (milking_update_new.trust_cal == 1),
            "calving",
        ] = milking_update_new.loc[
            (milking_update_new.parity_db == milking_update_new.parity)
            & (milking_update_new.calving_db > milking_update_new.calving_cal)
            & (milking_update_new.trust_cal == 1),
            "calving_cal",
        ].apply(
            pd.to_datetime
        )

        # if the parity is not the same, we assume that a new lactation was not inserted into the database
        milking_update_new.loc[
            (milking_update_new.parity_db != milking_update_new.parity), "calving"
        ] = milking_update_new.loc[
            (milking_update_new.parity_db != milking_update_new.parity), "calving_cal"
        ].apply(
            pd.to_datetime
        )

        milking_update_new.loc[
            pd.isnull(milking_update_new.calving), "calving"
        ] = milking_update_new.loc[
            pd.isnull(milking_update_new.calving), "calving_cal"
        ].apply(
            pd.to_datetime
        )

        # if the lactation_oid was not define, create a new lactation_id out of the unique animal_id and the parity
        milking_update_new.loc[
            pd.isnull(milking_update_new.lactation_oid), "lactation_oid"
        ] = milking_update_new.loc[
            pd.isnull(milking_update_new.lactation_oid), "animal_id"
        ].astype(
            "int64"
        ) * 10**9 + milking_update_new.loc[
            pd.isnull(milking_update_new.lactation_oid), "parity"
        ].astype(
            "int64"
        )

        # drop columns that were only used for calculations
        milking_update_new = milking_update_new.drop(
            columns=[
                "calving_cal",
                "parity_db",
                "lactation_oid_db",
                "calving_db",
                "gap",
                "periods",
                "trust_cal",
            ]
        )

        # Calculate new DIM
        milking_update_new.loc[
            ~pd.isnull(milking_update_new.ended_at)
            & ~pd.isnull(milking_update_new.calving),
            "dim",
        ] = (
            (
                milking_update_new.loc[
                    ~pd.isnull(milking_update_new.ended_at)
                    & ~pd.isnull(milking_update_new.calving),
                    "ended_at",
                ].apply(pd.to_datetime)
                - milking_update_new.loc[
                    ~pd.isnull(milking_update_new.ended_at)
                    & ~pd.isnull(milking_update_new.calving),
                    "calving",
                ].apply(pd.to_datetime)
            ).dt.total_seconds()
            / 86400
        ).round(
            decimals=2
        )

        # update lactation table
        lac_update = (
            milking_update_new[
                ["farm_id", "animal_id", "lactation_oid", "calving", "parity"]
            ]
            .drop_duplicates()
            .reset_index(drop=True)
        )
        lac_update.lactation_oid = lac_update.lactation_oid.astype(int)
        lac_update.loc[pd.isnull(lac_update.parity)] = 0
        lac_update.parity = lac_update.parity.astype(int)
        lac_update["calving"] = lac_update["calving"].apply(pd.to_datetime)
        lac_update["updated_on"] = dt.date.today()

        lac_update["calving"] = lac_update["calving"].dt.strftime("%Y-%m-%d %H:%M:%S")

        lactation_sql = create_sql_insert_update(
            sqlupdateDict["sql_update"]["milking_lactation_table"]
        )
        db_connect.insert(query=lactation_sql, data=lac_update.to_dict("records"))

        ################################################################################################################

        # add the lactation_ids back to the milking dataframe
        lactation_ids = db_connect.query(
            query=f"""SELECT lactation_id, lactation_oid FROM lactation WHERE farm_id = {farm_id};"""
        )

        dict_lactation_ids = pd.Series(
            lactation_ids.lactation_id.values, index=lactation_ids.lactation_oid
        ).to_dict()

        milking_update_new["lactation_id"] = milking_update_new["lactation_oid"].map(
            dict_lactation_ids
        )

        milking_update_new = milking_update_new.drop(columns=["lactation_oid"])

        # drop the data that was already available in the database
        milking_update_new["updated_on"] = dt.date.today()
        milking_update_db_data = (
            milking_update_new.loc[milking_update_new["_merge"] != "left_only"]
            .copy()
            .drop(columns=["_merge"])
        )
        milking_update_new = milking_update_new.loc[
            milking_update_new["_merge"] == "left_only"
        ].drop(columns=["_merge"])

        milking_update_new = milking_update_new.loc[
            ~pd.isnull(milking_update_new.milking_oid)
        ]
        milking_update_new = milking_update_new.loc[
            ~pd.isnull(milking_update_new.animal_id)
        ]
        milking_update_new = milking_update_new.loc[
            ~pd.isnull(milking_update_new.lactation_id)
        ]
        # milking_update_new = milking_update_new.loc[
        #     ~pd.isnull(milking_update_new.milking_system_oid)
        # ]

    ################################################################################################################

    if len(milking_update_db_data) > 0:
        # update milking table
        milking_sql = create_sql_update(
            sqlupdateDict["sql_update"]["milking_milking_table"]
        )

        db_connect.insert(
            query=milking_sql, data=milking_update_db_data.to_dict("records")
        )

    if len(milking_update_new) > 0:
        ################################################################################################################
        # if the milkings system oid is missing in the original backup data, create a new id, to indicate this and add the data regardsless to our database
        milking_update_new.loc[
            pd.isnull(milking_update_new.milking_system_oid), "milking_system_oid"
        ] = int(farm_id * 1000 + 999)

        ################################################################################################################
        # update milking system table
        milking_system_update = (
            milking_update_new[["farm_id", "milking_system_oid", "updated_on"]]
            .copy()
            .drop_duplicates(subset=["farm_id", "milking_system_oid"])
        )
        ms_update_sql = create_sql_insert_update(
            sqlupdateDict["sql_update"]["milking_system_milking_table"]
        )
        db_connect.insert(
            query=ms_update_sql, data=milking_system_update.to_dict("records")
        )

        ################################################################################################################
        # map milking system ids to milking system oids
        milking_system_ids = db_connect.query(
            query=f"""SELECT milking_system_id, milking_system_oid FROM milking_system WHERE farm_id = {farm_id};"""
        )

        dict_ms_ids = pd.Series(
            milking_system_ids.milking_system_id.values,
            index=milking_system_ids.milking_system_oid,
        ).to_dict()

        milking_update_new["milking_system_id"] = milking_update_new[
            "milking_system_oid"
        ].map(dict_ms_ids)

        milking_update_new = milking_update_new.drop(columns=["milking_system_oid"])

        ################################################################################################################

        milking_update_new = milking_update_new.drop(columns=["calving"])

        # insert new milkings into milking table
        datetime_columns = tableDict["milking"][f"{farmtype}_datetime"]
        for column in datetime_columns:
            milking_update_new[column] = milking_update_new[column].dt.strftime(
                "%Y-%m-%d %H:%M:%S"
            )
        milking_update_new.to_sql(
            "milking", con=db_connect.ret_con(), if_exists="append", index=False
        )


def mapMilkingSystem(df_data: pd.DataFrame, db_connect, farm_id: int) -> pd.DataFrame:
    """
    Update milking system table

    Parameters
    ----------
    df_data : pandas dataframe
        dataframe for a single table
    db_connect : sql engine
        engine to connect ot database
    farm_id : integer
        id of farm for which data is processed

    Return
    ------
    df_data : pandas dataframe
        dataframe with new milking_system_ids
    """
    milking_system_ids = db_connect.query(
        query=f"""SELECT milking_system_id, milking_system_oid FROM milking_system WHERE farm_id = {farm_id};"""
    )

    dict_ms_ids = pd.Series(
        milking_system_ids.milking_system_id.values,
        index=milking_system_ids.milking_system_oid,
    ).to_dict()

    ################################################################################################################
    # if the milkings system oid is missing in the original backup data, create a new id, to indicate this and add the data regardsless to our database
    df_data.loc[pd.isnull(df_data.milking_system_oid), "milking_system_oid"] = int(
        farm_id * 1000 + 999
    )

    df_data["milking_system_id"] = df_data["milking_system_oid"].map(dict_ms_ids)
    df_data.milking_system_id = df_data.milking_system_id.astype(float).astype('Int64')

    df_data = df_data.drop(columns=["milking_system_oid"])
    return df_data


def mapAnimalid(df_data: pd.DataFrame, db_connect, farm_id: int) -> pd.DataFrame:
    """
    Map new animal_ids to the animal_id used on the farm

    Parameters
    ----------
    df_data : pandas dataframe
        dataframe for a single table
    db_connect : sql engine
        engine to connect ot database
    farm_id : int
        id of farm for which data is processed

    Return
    ------
    df_data : pandas dataframe
        dataframe with new animal_ids
    """
    animal_ids = db_connect.query(
        query=f"""SELECT animal_id, animal_oid FROM animal WHERE farm_id = {farm_id};"""
    )

    dict_animal_ids = pd.Series(
        animal_ids.animal_id.values, index=animal_ids.animal_oid
    ).to_dict()

    df_data["animal_id"] = df_data["animal_oid"].map(dict_animal_ids)

    df_data = df_data.drop(columns=["animal_oid"])
    return df_data


def mapLactationid(
    df_data: pd.DataFrame, db_connect, farm_id: int, tablename: str
) -> pd.DataFrame:
    """
    Map new lactation_ids to the lactation_id used on the farm

    Parameters
    ----------
    df_data : pandas dataframe
        dataframe for a single table
    db_connect : sql engine
        engine to connect ot database
    farm_id : integer
        id of farm for which data is processed
    tablename : str
        name of the table/file

    Return
    ------
    df_data : pandas dataframe
        dataframe with new lactation_ids
    """
    lactation_sql = f"""
    SELECT farm_id, animal_id, lactation_id, calving, parity
    FROM lactation
    WHERE farm_id = {farm_id}
    ;
    """
    lactation_db = db_connect.query(query=lactation_sql)

    for lacids in list(
        lactation_db.sort_values(by=["animal_id", "calving"]).lactation_id
    ):
        aniid = lactation_db.loc[
            lactation_db.lactation_id == lacids, "animal_id"
        ].values[0]
        calvingdate = (
            lactation_db.loc[lactation_db.lactation_id == lacids, "calving"]
            .apply(pd.to_datetime)
            .values[0]
        )
        lacno = lactation_db.loc[lactation_db.lactation_id == lacids, "parity"].values[
            0
        ]
        df_data.loc[
            (df_data.animal_id == aniid) & (df_data.measured_on >= calvingdate),
            "lactation_id",
        ] = lacids
        df_data.loc[
            (df_data.animal_id == aniid) & (df_data.measured_on >= calvingdate),
            "calvingdate",
        ] = calvingdate
        df_data.loc[
            (df_data.animal_id == aniid) & (df_data.measured_on >= calvingdate),
            "parity",
        ] = lacno
    # calculate the dim
    df_data["dim"] = (
        df_data.measured_on - pd.to_datetime(df_data.calvingdate, format="%Y-%m-%d")
    ) / np.timedelta64(1, "D")
    if tablename in ["bcs", "activity"]:
        df_data = df_data.drop(columns=["calvingdate"])
    return df_data


def updateFarmLocation(db_connect, sqlupdateDict: dict, anonymized: bool) -> None:
    """
    Update farm location

    Parameters
    ----------
    db_connect : sql engine
        engine to connect ot database
    sqlupdateDict : dict
        dictionary containing information to generate a sql statement
    anonymized : bool
        anonlymize farm location (in a radius of 5 km around original location)
    """
    farm_sql = """
            SELECT *
            FROM farm
            ;
            """
    farm_db = db_connect.query(query=farm_sql)

    dict_farmlocation = sqlupdateDict["farminfo"]["farm_location"]

    dict_farm_lat = dict(
        zip(
            dict_farmlocation.keys(),
            list(list(zip(*list(dict_farmlocation.values())))[0]),
        )
    )
    dict_farm_long = dict(
        zip(
            dict_farmlocation.keys(),
            list(list(zip(*list(dict_farmlocation.values())))[1]),
        )
    )
    dict_farm_alt = dict(
        zip(
            dict_farmlocation.keys(),
            list(list(zip(*list(dict_farmlocation.values())))[2]),
        )
    )

    farm_db["longitude"] = farm_db["farmname"].map(dict_farm_long)
    farm_db["latitude"] = farm_db["farmname"].map(dict_farm_lat)
    farm_db["altitude"] = farm_db["farmname"].map(dict_farm_alt)

    if anonymized == True:
        for index, row in farm_db.iterrows():
            rand_dist = random.randint(0, 2 * np.pi)
            rand_long = np.sin(rand_dist) * 0.1
            rand_lat = np.cos(rand_dist) * 0.1
            farm_db.loc[farm_db.index == index, "longitude"] += rand_long
            farm_db.loc[farm_db.index == index, "latitude"] += rand_lat

    #########################################################################################################################
    # Update database table 'aws' with weather data

    farm_db["updated_on"] = dt.date.today()

    farm_update_sql = create_sql_update(sqlupdateDict["sql_update"]["farm_location"])
    db_connect.insert(query=farm_update_sql, data=farm_db.to_dict("records"))


def updateDatabaseTable(
    db_connect,
    filepath: str | Path,
    tablename: str,
    farmname: str,
    farmtype: str,
    tableDict: dict,
    sqlupdateDict: dict,
) -> None:
    """
    Update farm information

    Parameters
    ----------
    db_connect : sql engine
        engine to connect ot database
    filepath : str or Path
        path to datafile (.csv)
    tablename : str
        name of the table/file
    farmname : str
        name of farm for which data is processed
    farmtype : str
        milking system identifier (d- delaval, l- lely)
    tableDict : dict
        dictionary containing information about the table schema of the target database
    sqlupdateDict : dict
        dictionary containing information to generate a sql statement
    """

    if farmtype == "d":
        milkingSystem = "AMS delaval"
    elif farmtype == "l":
        milkingSystem = "AMS lely"

    df_data = readData(filepath, tablename, tableDict, farmtype)
    df_data, farm_id = updateFarm(
        df_data, db_connect, farmname, milkingSystem, sqlupdateDict
    )

    if tablename == "animal":
        df_data_update = checkDatabase(
            df_data, db_connect, tablename, farm_id, tableDict
        )
        if not df_data_update.empty:
            writeData(df_data_update, db_connect, tablename, tableDict, farmtype)
    elif tablename == "lactation":
        df_data = updateAnimal(df_data, db_connect, farm_id, sqlupdateDict)
        df_data = df_data.loc[~pd.isnull(df_data.animal_id)]
        df_data_update = checkDatabase(
            df_data, db_connect, tablename, farm_id, tableDict
        )
        if not df_data_update.empty:
            writeData(df_data_update, db_connect, tablename, tableDict, farmtype)
    elif tablename == "milking":
        df_data = updateAnimal(df_data, db_connect, farm_id, sqlupdateDict)
        updateMilking(
            df_data, db_connect, farm_id, farmname, farmtype, sqlupdateDict, tableDict
        )
    elif tablename == "cleaning":
        df_data = mapMilkingSystem(df_data, db_connect, farm_id)
        df_data_update = checkDatabase(
            df_data, db_connect, tablename, farm_id, tableDict
        )
        if not df_data_update.empty:
            writeData(df_data_update, db_connect, tablename, tableDict, farmtype)
    elif tablename in ["activity", "herdnavigator", "insemination", "bcs"]:
        df_data = mapAnimalid(df_data, db_connect, farm_id)
        df_data_update = checkDatabase(
            df_data, db_connect, tablename, farm_id, tableDict
        )
        if not df_data_update.empty:
            df_data_update = mapLactationid(
                df_data_update, db_connect, farm_id, tablename
            )
            writeData(df_data_update, db_connect, tablename, tableDict, farmtype)


def updateDatabaseFull(
    rootdir: str | Path, farm_to_update: str, anonymizeFarmlocation: bool
) -> None:
    """
    Update farm information

    Parameters
    ----------
    rootdir : str or Path
        path to cowbase rootdir
    farm_to_update : str
        name of farm for which data is processed
    anonymizeFarmlocation : bool
        anonymize farm location (in a radius of 5 km around original location)
    """

    rootdir = Path(rootdir)

    with open(rootdir / "config" / "serverSettings.json") as file:
        serverSettings = json.load(file)
    with open(rootdir / "config" / "M4_tableDict.json") as file:
        tableDict = json.load(file)
    with open(rootdir / "config" / "M4_sqlupdate.json") as file:
        sqlupdateDict = json.load(file)

    db_connect = DB_connect(**serverSettings)

    filepath = rootdir / "OUTPUT"
    farmfound = False
    for farms in os.listdir(filepath):
        farmtype = re.split("_", farms)[0]
        farmname = re.split("_", farms)[1]
        if not farms == farm_to_update:
            continue
        farmfound = True
        for tables in tableDict:
            for files in os.listdir(filepath / farms / "02_merged_table"):
                tablename = re.split("_", files[:-4])[-1]
                if tablename == tables:
                    print(f"Writing data in table {tablename}!")
                    datapath = filepath / farms / "02_merged_table" / files
                    updateDatabaseTable(
                        db_connect,
                        datapath,
                        tablename,
                        farmname,
                        farmtype,
                        tableDict,
                        sqlupdateDict,
                    )
                    print("Done!")
        updateFarmLocation(db_connect, sqlupdateDict, anonymizeFarmlocation)
    if farmfound == False:
        print("Could not find data for the farm that should be analyzed!")


def writeWeather(rootdir):
    rootdir = Path(rootdir)
    rootdir_INPUT_weather = rootdir / "INPUT" / "weather"
    with open(rootdir / "config" / "serverSettings.json") as file:
        serverSettings = json.load(file)
    with open(rootdir / "config" / "M4_sqlupdate.json") as file:
        sqlupdateDict = json.load(file)
    db_connect = DB_connect(**serverSettings)

    df_weather = pd.read_csv(rootdir_INPUT_weather / "weather.csv", header=0)
    df_weather_stations = pd.read_csv(rootdir_INPUT_weather / "aws.csv", header=0)
    df_weather_stations_weights = pd.read_csv(
        rootdir_INPUT_weather / "aws_weights.csv", header=0
    )
    df_weather_stations[
        [
            "hourly_start",
            "hourly_end",
            "daily_start",
            "daily_end",
            "monthly_start",
            "monthly_end",
        ]
    ] = df_weather_stations[
        [
            "hourly_start",
            "hourly_end",
            "daily_start",
            "daily_end",
            "monthly_start",
            "monthly_end",
        ]
    ].apply(
        pd.to_datetime
    )

    for column in [
        "hourly_start",
        "hourly_end",
        "daily_start",
        "daily_end",
        "monthly_start",
        "monthly_end",
    ]:
        df_weather_stations[column] = df_weather_stations[column].dt.strftime(
            "%Y-%m-%d %H:%M:%S"
        )

    df_weather_stations[
        [
            "hourly_start",
            "hourly_end",
            "daily_start",
            "daily_end",
            "monthly_start",
            "monthly_end",
        ]
    ] = df_weather_stations[
        [
            "hourly_start",
            "hourly_end",
            "daily_start",
            "daily_end",
            "monthly_start",
            "monthly_end",
        ]
    ].replace(
        {np.nan: None}
    )

    farm_sql = """
    SELECT farm_id, farmname
    FROM farm
    ;
    """
    farm_db = db_connect.query(query=farm_sql)

    dict_farm_ids = pd.Series(farm_db.farm_id.values, index=farm_db.farmname).to_dict()

    df_weather = df_weather.loc[df_weather.farmname.isin(list(farm_db.farmname))]
    df_weather_stations = df_weather_stations.loc[
        df_weather_stations.farmname.isin(list(farm_db.farmname))
    ]
    df_weather_stations_weights = df_weather_stations_weights.loc[
        df_weather_stations_weights.farmname.isin(list(farm_db.farmname))
    ]

    df_weather["farm_id"] = df_weather["farmname"].map(dict_farm_ids)
    df_weather = df_weather.drop(columns=["farmname"])
    df_weather["updated_on"] = dt.date.today()

    df_weather_stations = df_weather_stations.drop(columns=["farmname"])
    df_weather_stations["updated_on"] = dt.date.today()

    df_weather_stations_weights["farm_id"] = df_weather_stations_weights[
        "farmname"
    ].map(dict_farm_ids)
    df_weather_stations_weights = df_weather_stations_weights.drop(columns=["farmname"])
    df_weather_stations_weights["updated_on"] = dt.date.today()

    df_weather_stations_weights.aws_id_1 = np.floor(pd.to_numeric(df_weather_stations_weights.aws_id_1, errors='coerce')).astype('Int64')
    df_weather_stations_weights.aws_id_2 = np.floor(pd.to_numeric(df_weather_stations_weights.aws_id_2, errors='coerce')).astype('Int64')
    df_weather_stations_weights.aws_id_3 = np.floor(pd.to_numeric(df_weather_stations_weights.aws_id_3, errors='coerce')).astype('Int64')
    df_weather_stations_weights.aws_id_4 = np.floor(pd.to_numeric(df_weather_stations_weights.aws_id_4, errors='coerce')).astype('Int64')

    # ####################################################################################################################
    # drop data that was already in the database
    weather_sql = f"""
    SELECT farm_id, datetime
    FROM weather
    ;
    """
    weather_db = db_connect.query(query=weather_sql)

    df_weather["datetime"] = df_weather["datetime"].apply(pd.to_datetime)
    weather_db["datetime"] = weather_db["datetime"].apply(pd.to_datetime)

    df_weather = df_weather.drop_duplicates(
        subset=["farm_id", "datetime"], keep="first"
    )

    weather_update_new = df_weather.merge(
        weather_db[["farm_id", "datetime"]],
        how="left",
        on=["farm_id", "datetime"],
        indicator=True,
    )

    weather_update_new = weather_update_new.loc[
        weather_update_new["_merge"] == "left_only"
    ].drop(columns=["_merge"])

    weather_update_new["datetime"] = weather_update_new["datetime"].dt.strftime(
        "%Y-%m-%d %H:%M:%S"
    )

    weather_update_new.to_sql(
        "weather", con=db_connect.ret_con(), if_exists="append", index=False
    )

    ####################################################################################################################

    aws_update_sql = create_sql_insert_update(sqlupdateDict["sql_update"]["aws"])
    db_connect.insert(query=aws_update_sql, data=df_weather_stations.to_dict("records"))

    aws_weights_update_sql = create_sql_insert_update(
        sqlupdateDict["sql_update"]["aws_weights"]
    )
    db_connect.insert(
        query=aws_weights_update_sql,
        data=df_weather_stations_weights.to_dict("records"),
    )


def meteostatExtract(rootdir, startdate="2010-01-01", enddate="2023-01-01"):
    start = dt.datetime.strptime(startdate, "%Y-%m-%d")
    end = dt.datetime.strptime(enddate, "%Y-%m-%d")

    with open(os.path.join(rootdir, "config", "M4_sqlupdate.json")) as file:
        sqlupdateDict = json.load(file)

    dict_farmlocation = sqlupdateDict["farminfo"]["farm_location"]

    weather = pd.DataFrame()
    weather_stations = pd.DataFrame()
    weather_stations_weights = pd.DataFrame()
    for farms in dict_farmlocation.keys():
        print(farms)
        farmlocation = Point(
            dict_farmlocation[farms][0],
            dict_farmlocation[farms][1],
            dict_farmlocation[farms][2],
        )
        farmlocation.method = "weighted"
        weather_add = Hourly(farmlocation, start, end)
        weather_add = weather_add.fetch()
        weather_add = weather_add.reset_index()
        weather_add = weather_add.rename(
            columns={
                "time": "datetime",
                "temp": "temperature",
                "dwpt": "dew_point",
                "rhum": "humidity",
                "prcp": "precipitation",
                "snow": "snow",
                "wdir": "wind_direction",
                "wspd": "windspeed",
                "wpgt": "peak_wind_gust",
                "pres": "air_pressure",
                "tsun": "total_sunshine_duration",
                "coco": "weather_condition_code",
            }
        )
        weather_stations_add = Stations()
        weather_stations_add = weather_stations_add.nearby(
            dict_farmlocation[farms][0], dict_farmlocation[farms][1]
        ).fetch(4)
        weather_stations_add = weather_stations_add.reset_index()
        weather_stations_add = weather_stations_add.rename(columns={"id": "aws_id"})
        weather_stations_add = weather_stations_add.loc[~pd.isnull(weather_stations_add.wmo)]
        weather_stations_add = weather_stations_add.reset_index(drop=True)
        weather_stations_add.aws_id = weather_stations_add.aws_id.astype(int)
        if len(weather_stations_add) == 4:
            weather_stations_weighted_add = pd.DataFrame(
                {
                    "farmname": [farms],
                    "aws_id_1": list(weather_stations_add.aws_id.astype(int))[0],
                    "distance_1": list(weather_stations_add.distance.astype(float))[0],
                    "aws_id_2": list(weather_stations_add.aws_id.astype(int))[1],
                    "distance_2": list(weather_stations_add.distance.astype(float))[1],
                    "aws_id_3": list(weather_stations_add.aws_id.astype(int))[2],
                    "distance_3": list(weather_stations_add.distance.astype(float))[2],
                    "aws_id_4": list(weather_stations_add.aws_id.astype(int))[3],
                    "distance_4": list(weather_stations_add.distance.astype(float))[3],
                }
            )
        elif len(weather_stations_add) == 3:
            print('Warning: 1 of the closest weather stations were not registered with WMO and hence removed!')
            weather_stations_weighted_add = pd.DataFrame(
                {
                    "farmname": [farms],
                    "aws_id_1": list(weather_stations_add.aws_id.astype(int))[0],
                    "distance_1": list(weather_stations_add.distance.astype(float))[0],
                    "aws_id_2": list(weather_stations_add.aws_id.astype(int))[1],
                    "distance_2": list(weather_stations_add.distance.astype(float))[1],
                    "aws_id_3": list(weather_stations_add.aws_id.astype(int))[2],
                    "distance_3": list(weather_stations_add.distance.astype(float))[2],
                }
            )
        elif len(weather_stations_add) == 2:
            print('Warning: 2 of the closest weather stations were not registered with WMO and hence removed!')
            weather_stations_weighted_add = pd.DataFrame(
                {
                    "farmname": [farms],
                    "aws_id_1": list(weather_stations_add.aws_id.astype(int))[0],
                    "distance_1": list(weather_stations_add.distance.astype(float))[0],
                    "aws_id_2": list(weather_stations_add.aws_id.astype(int))[1],
                    "distance_2": list(weather_stations_add.distance.astype(float))[1],
                    "aws_id_3": list(weather_stations_add.aws_id.astype(int))[2],
                    "distance_3": list(weather_stations_add.distance.astype(float))[2],
                }
            )
        else:
            print('Warning: More than 2 of the closest weather stations were not registered with WMO and hence removed!')
            weather_stations_weighted_add = pd.DataFrame()

        weather_stations_add = weather_stations_add.drop(columns=["distance"])
        weather_stations_add["farmname"] = farms
        weather_add["farmname"] = farms
        weather = pd.concat([weather, weather_add])
        weather = weather.reset_index(drop=True)
        weather_stations = pd.concat([weather_stations, weather_stations_add])
        weather_stations = weather_stations.drop_duplicates(subset=["aws_id"])
        weather_stations = weather_stations.reset_index(drop=True).sort_values(
            by="aws_id"
        )
        weather_stations_weights = pd.concat(
            [weather_stations_weights, weather_stations_weighted_add]
        )
        weather_stations_weights = weather_stations_weights.reset_index(drop=True)

    rootdir = Path(rootdir)
    rootdir_INPUT_weather = rootdir / "INPUT" / "weather"
    rootdir_INPUT_weather.mkdir(exist_ok=True, parents=True)
    weather.to_csv(rootdir_INPUT_weather / "weather.csv", index=False)
    weather_stations.to_csv(rootdir_INPUT_weather / "aws.csv", index=False)
    weather_stations_weights.to_csv(
        rootdir_INPUT_weather / "aws_weights.csv", index=False
    )



import json
import os
import platform
import re
import shutil
import zipfile
from datetime import datetime
from pathlib import Path

import pandas as pd
import tomli


def initCowBase(rootdir: str) -> None:
    """
    Initializes the folder structure in which the CowBase package will be mainly operating.

    Parameters
    ----------
    rootdir : str or Path
        Root directory in which the CowBase folder structure should be implemented in.

    Notes
    ----------
    A number of folders and files is created when initializing:
    1. CowBase - parent directory
    2. CowBase/INPUT - folder in which the AMS (automatic milking system) backup files (.bak.zip/.zip) should be saved
    3. CowBase/INPUT/d_farmname - Backups from DeLaval systems should always be saved in folders starting with 'd_' and followed by the farmname
    4. CowBase/INPUT/l_farmname - Backups from Lely systems should always be saved in folders starting with 'l_' and followed by the farmname
    5. CowBase/OUTPUT - folder in which the unpacked data files (.csv) will be later saved
    6. CowBase/temp - folder in which temparary files are stored
    7. CowBase/config - folder in which a user config file and dictionaries are stored

    """

    rootdir = Path(rootdir)

    # initialize folder structure
    rootdir_cowbase = rootdir / "CowBase"
    rootdir_cowbase.mkdir(exist_ok=True, parents=True)

    rootdir_INPUT = rootdir_cowbase / "INPUT"
    rootdir_INPUT.mkdir(exist_ok=True, parents=True)

    rootdir_INPUT_dtest = rootdir_INPUT / "d_farmname"
    rootdir_INPUT_dtest.mkdir(exist_ok=True)

    rootdir_INPUT_ltest = rootdir_INPUT / "l_farmname"
    rootdir_INPUT_ltest.mkdir(exist_ok=True)

    rootdir_OUTPUT = rootdir_cowbase / "OUTPUT"
    rootdir_OUTPUT.mkdir(exist_ok=True)

    rootdir_temp = rootdir_cowbase / "temp"
    rootdir_temp.mkdir(exist_ok=True)

    rootdir_config = rootdir_cowbase / "config"
    rootdir_config.mkdir(exist_ok=True)

    # add config and tableDict to the config folder for individual settings
    for config_file in (Path(__file__).parent / "utils").glob("*"):
        # If the config file is already in dir, don't copy again
        if (rootdir_config / config_file.name).exists():
            continue

        shutil.copy(
            config_file,
            rootdir_config / config_file.name,
        )

    # set the rootdir in the config file to the defined rootdir
    # TODO: if the config file has already been edited, don't overwrite
    osname = os.name
    fin = open((Path(__file__).parent / "utils" / "config.toml"), "rt")
    fout = open(f"{rootdir_config / 'config.toml'}", "wt")
    for line in fin:
        # read replace the string and write to output file
        if osname == "nt":
            hostname = platform.node()
            fout.write(
                line.replace("SERVERNAME", f"{hostname}\SQLEXPRESS").replace(
                    "rootdir_cowbase", str(rootdir_cowbase)
                )
            )
        elif osname == "posix":
            fout.write(
                line.replace("SERVERNAME", "localhost").replace(
                    "rootdir_cowbase", str(rootdir_cowbase)
                )
            )
    # close input and output files
    fin.close()
    fout.close()


def extracttable(
    tabledict: dict,
    table: str,
    robottype: str,
    dbname: str,
    filename_out: str,
    config: dict,
) -> None:
    """
    Extracts data tables from restored MSQL database

    Parameters
    ----------
    tabledict : dict
        dictonary that contains all table and column names from the respective milking system that should be extracted
    table : str
        tablename of the table that should be restored
    robottype : str
        brand of milking system the backup is from (d-DeLaval; l-Lely)
    dbname : str
        name of the database
    filename_out : str or Path
        filepath and name to which the extracted data will be saved to
    config : dict
        config file with user settings

    """

    osname = os.name
    columnnames = ", ".join(tabledict[robottype][table])
    # export table
    sql = f"SELECT {columnnames} FROM {dbname}.dbo.{table}"
    servername = config["user"]["servername"]
    if osname == "nt":
        cmd = f'bcp "{sql}" queryout {filename_out} -d{dbname} -c -C 65001 -t"^^" -m 9999  -b 10000 -a 10000 -k -T -S {servername}'
    elif osname == "posix":
        pwd = config["user"]["serverpwd"]
        cmd = f'bcp "{sql}" queryout {filename_out} -d {dbname} -c -u -C 65001 -t"^^" -m 9999  -b 10000 -a 10000 -k -U sa -P "{pwd}" -S {servername}'
    else:
        print(f"CowBase is not yet adapted to your operating system ({os.name}).")
        return

    os.system(cmd)
    try:
        # with open(filename_out, "rb") as source_file:
        #     with open(filename_out, "wb") as dest_file:
        #         contents = source_file.read()
        #         dest_file.write(contents.replace(b"\x00", b""))
        datatable = pd.read_csv(
            filename_out, header=None, delimiter="\^\^", engine="python"
        )
        datatable = datatable.replace({";": ","}, regex=True)
        datatable.to_csv(
            filename_out,
            header=tabledict[robottype][table],
            index=False,
            sep=";",
            escapechar="\\",
        )
    except pd.errors.EmptyDataError:
        os.remove(filename_out)
        print(f"File removed! No data available for {table}!")


def restoreDB_single(
    rootdir: str = "/var/lib/CowBase", farm: str = "d_farmname"
) -> None:
    """
    1. Restores MSQL database from backup-dump file (.bak)
    2. Calls extracttable() - extracts data tables defined in tabelDict.json (located in the config folder after initializing the pipeline)
    3. Drops the database and deletes temporary files

    Parameters
    ----------
    rootdir : str or Path
        Filepath to CowBase root directory
    farm : str
        Name of the single farm-directory for which data should be restored

    Note
    ----------
    Currently only backups from milking systems form DeLaval or Lely are supported.\n
    Further, currently only posix and windows nt based systems are supported.

    """
    rootdir = Path(rootdir)

    filepath_config = rootdir / "config" / "config.toml"
    filepath_tdict = rootdir / "config" / "M1_tableDict.json"
    with open(filepath_config, mode="rb") as ctomli:
        config = tomli.load(ctomli)
    osname = os.name

    # Specify the folder where the backups are placed and create a temp folder to temporary store the unzipped backups
    # as well as the OUTPUT folder where the extracted data gets stored
    with open(filepath_tdict) as file:
        tabledict = json.load(file)

    sql_template = rootdir / "config" / "M1_restoreDB.sql"
    config_user_rootdir = Path(config["user"]["rootdir"])

    rootdir_temp = config_user_rootdir / "temp"
    rootdir_temp.mkdir(exist_ok=True)

    rootdir_in = config_user_rootdir / "INPUT"
    rootdir_in.mkdir(exist_ok=True)

    rootdir_out = config_user_rootdir / "OUTPUT"
    rootdir_out.mkdir(exist_ok=True)

    # Check for backup files in folder farm and start extraction process
    print(f"Extracting data for farm {farm}!")
    filepath_in = rootdir_in / farm
    filepath_out = rootdir_out / farm / "01_full_table"
    filepath_out.mkdir(exist_ok=True, parents=True)

    for foldername in os.listdir(filepath_in):
        filetype = None
        filedate = None
        # Lely and DeLaval have .zip/.bak.zip directories for their backup files
        if ".zip" in foldername:
            filetype = "zip"
            filename = foldername.replace(".zip", "")
        if ".bak" in foldername:
            filetype = "bak" + filetype
            filename = foldername.replace(".bak", "")
        # Restore database for backup files/folders
        if filetype == "zip" or filetype == "bakzip":
            print(f"Extracting data for backup {filename}!")
            robottype = farm[0:1]

            # Search for robotversion and file date in file string
            robotversion_searchstr = re.search(r"DelPro+[3-6]+\.[0-9]{1,2}", filename)
            filedate_searchstr = re.search(
                r"20+[0-2]+[0-9]+[0-1]+[0-9]+[0-3]+[0-9]", filename
            )
            if robotversion_searchstr != None:
                robotversion = float(robotversion_searchstr.group()[6:])
            else:
                robotversion = 1.0
            if filedate_searchstr != None:
                filedate = filedate_searchstr.group()
            else:
                filedate = datetime.now().strftime("%Y%m%d")
            # For DeLaval database version 3.7, there is a seperate, slightly different database scheme available
            if robotversion == 3.7:
                robottype = "d_37"

            # Unzip file to temp folder
            with zipfile.ZipFile(filepath_in / foldername, "r") as zip:
                files = zip.namelist()
                for file in files:
                    filename, extension = os.path.splitext(file)
                    if extension == ".bak":
                        zip.extract(file, rootdir_temp)

            backup_string = ""
            num_bfiles = 0
            # TODO: I think this should only loop on .bak files
            for bfiles in os.listdir(rootdir_temp):
                num_bfiles += 1
                backup_string = (
                    backup_string + f"DISK = N''' + @DirOutput + '{bfiles}'', "
                )
            backup_1 = os.listdir(rootdir_temp)[0]
            backup_string = backup_string[:-2]

            # Check that there is only one backup file, if corrupted and the backup file is split, return with an error message
            if num_bfiles == 1:
                fin = open(sql_template, "rt")
                fout = open(f"{rootdir_temp / 'restoreDB.sql'}", "wt")
                for line in fin:
                    # read replace the string and write to output file
                    fout.write(
                        line.replace("VARTEMPDIR", str(rootdir_temp) + os.sep)
                        .replace("VARBACKUPNAMES", backup_string)
                        .replace("VARBACKUPNAME1", backup_1)
                    )
                # close input and output files
                fin.close()
                fout.close()

                # Restore database
                if osname == "nt":
                    cmd = f'''sqlcmd -S {config['user']['servername']} -E -i "{rootdir_temp / 'restoreDB.sql'}" -o "{rootdir_temp / 'sqlout.txt'}"'''
                elif osname == "posix":
                    cmd = f'''sqlcmd -S {config['user']['servername']} -U SA -P '{config['user']['serverpwd']}' -C -i "{rootdir_temp / 'restoreDB.sql'}" -o "{rootdir_temp / 'sqlout.txt'}"'''
                else:
                    print(
                        f"CowBase is not yet adapted to your operating system ({os.name})."
                    )
                    return
                os.system(cmd)

                with open(rootdir_temp / "sqlout.txt") as f:
                    lines = f.readlines()
                list_sqlout = [re.sub(r"[\n\t\s]*", "", i) for i in lines]
                dbname = list_sqlout[list_sqlout.index("dbname") + 1]
                dblog = list_sqlout[list_sqlout.index("dblog") + 1]

                # print("\n\n***DEBUG: DBNAME = ", dbname, "\n\n")

                # For all data tables, query database and extract according to table dictionary (tabledict)
                for table in tabledict[robottype]:
                    if robotversion == 3.7:
                        filepath_out2 = filepath_out / (filedate + "_37")
                    else:
                        filepath_out2 = filepath_out / filedate

                    filepath_out2.mkdir(exist_ok=True, parents=True)
                    filename_out = filepath_out2 / f"{farm}_{filedate}_{table}.csv"
                    extracttable(
                        tabledict, table, robottype, dbname, filename_out, config
                    )
            else:
                print("Error: Backup is corrupted and can not be restored!")

            # Drop any database if still running
            if osname == "nt":
                cmd = f'''sqlcmd -S {config['user']['servername']} -E -Q "IF DB_ID ('DDM') IS NOT NULL DROP DATABASE DDM;"'''
            elif osname == "posix":
                cmd = f'''sqlcmd -S {config['user']['servername']} -U SA -P '{config['user']['serverpwd']}' -C -Q "IF DB_ID ('DDM') IS NOT NULL DROP DATABASE DDM;"'''
            else:
                print(
                    f"CowBase is not yet adapted to your operating system ({os.name})."
                )
                return
            os.system(cmd)

            if osname == "nt":
                cmd = f'''sqlcmd -S {config['user']['servername']} -E -Q "IF DB_ID ('DDMVMS') IS NOT NULL DROP DATABASE DDMVMS;"'''
            elif osname == "posix":
                cmd = f'''sqlcmd -S {config['user']['servername']} -U SA -P '{config['user']['serverpwd']}' -C -Q "IF DB_ID ('DDMVMS') IS NOT NULL DROP DATABASE DDMVMS;"'''
            else:
                print(
                    f"CowBase is not yet adapted to your operating system ({os.name})."
                )
                return
            os.system(cmd)

            if osname == "nt":
                cmd = f'''sqlcmd -S {config['user']['servername']} -E -Q "IF DB_ID ('LELY_Data') IS NOT NULL DROP DATABASE LELY_Data;"'''
            elif osname == "posix":
                cmd = f'''sqlcmd -S {config['user']['servername']} -U SA -P '{config['user']['serverpwd']}' -C -Q "IF DB_ID ('LELY_Data') IS NOT NULL DROP DATABASE LELY_Data;"'''
            else:
                print(
                    f"CowBase is not yet adapted to your operating system ({os.name})."
                )
                return
            os.system(cmd)

            # remove the .bak files from temporary storage
            for items in os.listdir(rootdir_temp):
                os.remove(os.path.join(rootdir_temp, items))


def restoreDB_all(rootdir: str = "/var/lib/CowBase") -> None:
    """
    Extracts data tables from restored MSQL database

    Parameters
    ----------
    rootdir : str or Path
        Filepath to CowBase root directory (e.g. linux: '/var/lib/'; windows: 'C:\\Users\\user\\Documents\\')


    Input (config.toml)
    -----
    servername : str
        # e.g. linux: 'localhost'; windows: 'PC-name\SQLEXPRESS' (PC-name can be found by typing 'hostname' in cmd)
    rootdir : str
        Filepath to CowBase root directory (e.g. linux: '/var/lib/'; windows: 'C:\\Users\\user\\Documents\\')
    serverpwd : str
        # usually only necessary for linux; password you set when creating the MSQL server

    """

    rootdir = Path(rootdir)

    filepath_config = rootdir / "config" / "config.toml"
    with open(filepath_config, mode="rb") as ctomli:
        config = tomli.load(ctomli)

    config_user_rootdir = Path(config["user"]["rootdir"])
    rootdir_temp = config_user_rootdir / "temp"
    rootdir_temp.mkdir(exist_ok=True)

    rootdir_in = config_user_rootdir / "INPUT"
    rootdir_in.mkdir(exist_ok=True)

    # Iterate all folders in the rootdir and check if there are any .zip or .bak.zip files
    for folder in os.listdir(rootdir_in):
        if folder[0] == 'd' or folder[0] == 'l':
            restoreDB_single(rootdir=rootdir, farm=folder)
        else:
            print(f'The folder {folder} is not in the right format. Please make sure the AMS type (l = Lely, d = DeLaval) was added to the farmname (d_farmname)!')



import json
import os
import pandas as pd
import numpy as np
from sshtunnel import SSHTunnelForwarder
from sqlalchemy import create_engine
from sqlalchemy import inspect
from sqlalchemy import text
from sqlalchemy_utils import database_exists, create_database
from sqlalchemy.ext.declarative import declarative_base
import pandas as pd
import os
import json

# import getpass


class DB_connect(object):

    """
    The class 'LT_connect' can be used to connect to the LT postgres database
    via a ssh connection.\n

    Functions:
    ----------

    tables(self, db, schema, sql_user, sql_pass)\n
    print_columns(self, db, table_name, sql_user, sql_pass)\n
    query(self, db, query, sql_user, sql_pass)\n
    execute(self, db, query, sql_user, sql_pass)\n
    insert(self, db, query, data, sql_user, sql_pass)\n
    ret_con(self, db, sql_user, sql_pass)\n
    create_db(self, db, sql_user, sql_pass)

    Parameters
    ----------

    db : name of a postgres database that should be accessed \n
    p_host : address of the database of the system
    (usually localhost - 127.0.0.1) \n
    p_port : port for postgresql (usually 5432) \n
    ssh : if a ssh connection is necessary insert 'True' \n
    ssh_user : account name of the ssh user \n
    ssh_host : ip address of the server to which to connect \n
    ssh_pkey : filepath to the ssh key for faster access \n
    sql_user : account name of the postgres user \n
    sql_pass : password for the postgres account \n

    Return
    ------

    None


    """

    def __init__(
        self,
        ssh,
        ssh_host,
        ssh_port,
        ssh_user,
        keybased,
        ssh_pwd,
        ssh_pkey,
        db_host,
        db_port,
        db,
        sql_user,
        sql_pass,
        dbtype,
        sqlitepath,
    ):
        """
        __init__(self, db_host, db_port, db, ssh, ssh_user, ssh_host, ssh_pkey, sql_user, sql_pass):
        -----------------------------------------------
        defines global class parameters for ssh connection\n

        Parameters
        ----------
        ssh : if a ssh connection is necessary insert 'True' \n
        ssh_host : ip address of the server to which to connect \n
        ssh_port : port of the server to which to connect \n
        ssh_user : account name of the ssh user \n
        keybased : boolean - True if ssh-key is used to connect to server \n
        ssh_pwd : password for ssh connection \n
        ssh_pkey : filepath to the ssh key for faster access \n
        db_host : address of the database of the system
        (usually localhost - 127.0.0.1) \n
        db_port : port for postgresql (usually 5432) \n
        db : name of a postgres database that should be accessed \n
        sql_user : account name of the postgres user \n
        sql_pass : password for the postgres account \n
        dbtype : dbtype used (postgres, mysql, sqlite)

        Returns:
        --------
        None
        """
        # SSH Tunnel Variables
        self.db_host = db_host
        self.db_port = db_port
        self.sql_user = sql_user
        self.sql_pass = sql_pass
        self.ssh_port = ssh_port
        self.dbtype = dbtype
        self.db = db
        self.echoparam = False
        
        if ssh == True:
            if keybased == True:
                self.server = SSHTunnelForwarder(
                    (ssh_host, ssh_port),
                    ssh_username=ssh_user,
                    ssh_pkey=ssh_pkey,
                    remote_bind_address=(db_host, db_port),
                )
                server = self.server
                server.start()  # start ssh server
                self.local_port = server.local_bind_port
                print(f"Server connected via SSH ...")
            else:
                self.server = SSHTunnelForwarder(
                    (ssh_host, ssh_port),
                    ssh_username=ssh_user,
                    ssh_password=ssh_pwd,
                    remote_bind_address=(db_host, db_port),
                )
                server = self.server
                server.start()  # start ssh server
                self.local_port = server.local_bind_port
                print(f"Server connected via SSH ...")
        else:
            self.local_port = db_port
        
        if dbtype == "postgres":
            self.enginestr = f"postgresql://{self.sql_user}:{self.sql_pass}@{self.db_host}:{self.local_port}/{self.db}"

        elif dbtype == "mysql":
            self.enginestr = f"mysql+mysqldb://{self.sql_user}:{self.sql_pass}@{self.db_host}:{self.local_port}/{self.db}"

        elif dbtype == "sqlite":
            if not os.path.exists(sqlitepath):
                os.makedirs(sqlitepath)
            self.enginestr = f"sqlite:///{sqlitepath}\\{self.db}.db"

    def tables(self, schema):
        """
        tables(self, db, schema, sql_user, sql_pass):
        -----------------------------------------------
        returns all table names in a given 'schema' of a database 'db'\n

        Parameters:
        ----------

        db : name of a postgres database that should be accessed \n
        schema : name of the schema that should be analyzed\n
        sql_user : account name of the postgres user \n
        sql_pass : password for the postgres account \n

        Returns:
        --------
        tables_df (pandas dataframe of table names)

        """

        # create an engine to connect to the postgreSQL database, documentation
        # of functions can be found at https://docs.sqlalchemy.org/en/14/

        if self.dbtype == "sqlite":
            engine = create_engine(self.enginestr, echo=True)
        else:
            engine = create_engine(self.enginestr)
        conn = engine.connect() 
        inspector = inspect(engine)
        tables = inspector.get_table_names(schema=schema)
        self.tables_df = pd.DataFrame(tables, columns=["table name"])
        engine.dispose()
        return self.tables_df

    def print_columns(self, table_name):
        """
        print_columns(self, db, table_name, sql_user, sql_pass)
        -----------------------------------------------
        returns all table names in a given 'schema' of a database 'db'\n

        Parameters:
        ----------

        db : name of a postgres database that should be accessed \n
        table_name : name of the table for which the columns schould be checked \n
        sql_user : account name of the postgres user \n
        sql_pass : password for the postgres account \n

        Returns:
        --------
        tables_df (pandas dataframe of column names)

        """

        # create an engine to connect to the postgreSQL database, documentation
        # of functions can be found at https://docs.sqlalchemy.org/en/14/

        engine = create_engine(self.enginestr, echo=self.echoparam)

        if " " in table_name:
            if '"' in table_name:
                pass
            else:
                table_name = "'" + table_name + "'"
        query = (
            """
        SELECT column_name, data_type
        FROM information_schema.columns
        WHERE table_name = %s
        ;
        """
            % table_name
        )
        self.table_df = pd.read_sql(query, engine)
        engine.dispose()
        return self.table_df

    def query(self, query):
        """
        query(self, db, query, sql_user, sql_pass)
        -----------------------------------------------
        executes a postgreSQL query in the database 'db' (return = true)\n

        Parameters:
        ----------

        db : name of a postgres database that should be accessed \n
        query : insert char string of postgreSQL code that should be queried \n
        sql_user : account name of the postgres user \n
        sql_pass : password for the postgres account \n

        Returns:
        --------
        query_df (pandas dataframe of query result)

        """
        # create an engine to connect to the postgreSQL database, documentation
        # of functions can be found at https://docs.sqlalchemy.org/en/14/
        engine = create_engine(self.enginestr, echo=self.echoparam)
        self.query_df = pd.read_sql(query, engine)
        engine.dispose()
        return self.query_df

    def execute(self, query):
        """
        execute(self, db, query, sql_user, sql_pass)
        -----------------------------------------------
        executes a postgreSQL query in the database 'db' (return = false)\n

        Parameters:
        ----------

        db : name of a postgres database that should be accessed \n
        query : insert char string of postgreSQL code that should be queried \n
        sql_user : account name of the postgres user \n
        sql_pass : password for the postgres account \n

        Returns:
        --------
        None

        """
        # create an engine to connect to the postgreSQL database, documentation
        # of functions can be found at https://docs.sqlalchemy.org/en/14/
        engine = create_engine(self.enginestr, echo=self.echoparam)
        with engine.begin() as connection:
            connection.execute(text(query))
        # engine.execute(text(query))
        engine.dispose()

    def insert(self, query, data):
        """
        insert(self, db, query, data, sql_user, sql_pass)
        -----------------------------------------------
        executes a postgreSQL query in the database 'db' (return = false),
        used to insert data with parameter data, use '%(name)s' in the query text
        and a dictionary ({name : value}) for data \n

        Parameters:
        ----------

        db : name of a postgres database that should be accessed \n
        query : insert char string of postgreSQL code that should be queried \n
        data : dictionary of data that should be used in the query \n
        sql_user : account name of the postgres user \n
        sql_pass : password for the postgres account \n

        Returns:
        --------
        None

        """
        # create an engine to connect to the postgreSQL database, documentation
        # of functions can be found at https://docs.sqlalchemy.org/en/14/
        engine = create_engine(self.enginestr, echo=self.echoparam)

        with engine.begin() as connection:
            connection.execute(text(query), data)
        # engine.execute(text(query), data[0])
        engine.dispose()

    def ret_con(self):
        """
        ret_con(self, db, sql_user, sql_pass)
        -----------------------------------------------
        returns the engine to connect to the database 'db'\n

        Parameters:
        ----------

        db : name of a postgres database that should be accessed \n
        sql_user : account name of the postgres user \n
        sql_pass : password for the postgres account \n

        Returns:
        --------
        engine

        """
        # create an engine to connect to the postgreSQL database, documentation
        # of functions can be found at https://docs.sqlalchemy.org/en/14/
        engine = create_engine(self.enginestr, echo=self.echoparam)
        return engine

    def create_db(self):
        """
        create_db(self, db, sql_user, sql_pass)
        -----------------------------------------------
        creates the database 'db'\n

        Parameters:
        ----------

        db : name of a postgres database that should be accessed \n
        sql_user : account name of the postgres user \n
        sql_pass : password for the postgres account \n

        Returns:
        --------
        None

        """
        # create an engine to connect to the postgreSQL database, documentation
        # of functions can be found at https://docs.sqlalchemy.org/en/14/

        print(self.dbtype)
        if self.dbtype == "sqlite":
            engine = create_engine(self.enginestr, echo=self.echoparam)
            Base = declarative_base()
            Base.metadata.create_all(engine)
        else:
            engine = create_engine(self.enginestr)
            if not database_exists(engine.url):
                create_database(engine.url)
            else:
                print('A database with the name "' + self.db + '" already exists')


def connCowBaseDB(rootdir):
    """
    Returns connection to database, specify parameters in settings file

    Parameters
    ----------
    rootdir : str
        path to cowbase rootdir
    """

    settingsPath = os.path.join(rootdir, "config", "serverSettings.json")

    with open(settingsPath) as file:
        serverSettings = json.load(file)

    # Create a connection to the database
    db_connect = DB_connect(**serverSettings)

    return db_connect


def dailyMY(df_milking):
    """
    Converts raw SESSION milk yield data to standardized DAILY milk yield data
    + some data selection steps (outliers deleted)

    Parameters
    ----------
    df_milking : df
        raw data of one farm from the database containing variables
            * milking_id
            * farm_id
            * animal_id
            * lactation_id
            * parity
            * started_at
            * ended_at
            * mi
            * dim
            * tmy & qmy (mylf, myrf, mylr, myrr)
            * quarter ec (eclf, ecrf, eclr, ecrr)

    Yields
    ------
    df_milking_daily : df
        (daily milk yield) data of one farm containing variables
            * farm_id
            * animal_id
            * lactation_id
            * dim
            * tdmy
            * dmylf, dmyrf, dmylr, dmyrr

    Steps
    -----
        - calculate the proportion of the MI before and after midnight
        - assign that proportion of the MY to the current or previous day (at udder and quarter level)
        - sum the milk yields of each day
        - add 1 to the DIM to avoid problems with log later on
        - standardize by dividing by the sum of all MI assigned to the current day (to account for missing milkings)
        - calculate a 7 day rolling median of the TDMY: delete tdmy values that are lower than 1.5*the rolling median
        - delete observations with TDMY below 75 kg

    """

    # Kick out all entries where mi could not be calculated/was not recorded.
    # This might lead to loss of information, creating a
    # total daily milk yield though would have large deviations if the mi was not guessed correctly.
    df_milking = df_milking.sort_values(
        by=["farm_id", "animal_id", "lactation_id", "milking_id"]
    ).reset_index(drop=True)
    milking = df_milking[df_milking["mi"].notna()].copy()

    # Floor dim for the calculation of daily milk yield
    milking["dim_floor"] = milking["dim"].apply(np.floor)

    # IV_dim_floor gives the time in hours at which the milking started at
    milking["IV_dim_floor"] = 24 * (milking["dim"] - milking["dim_floor"])

    # calculate the time between milkings that was during the previous day (compared to the dim floored)
    milking["mi_day_before"] = milking["mi"] - milking["IV_dim_floor"]
    # you put it to 0 because there was a previous milking on the same day
    milking.loc[milking["mi_day_before"] < 0, "mi_day_before"] = 0

    # calculate the time between milkings that was during the day (dim floored)
    milking["mi_on_day"] = milking["mi"] - milking["mi_day_before"]

    # calulate the proportions of mi on the day and day before
    milking["mi_day_before"] = milking["mi_day_before"] / milking["mi"]
    milking["mi_on_day"] = milking["mi_on_day"] / milking["mi"]

    # create a new table where the time between milkings was spread over two days
    MY_daily_add = milking[milking["mi_day_before"] > 0].copy()

    # multiply the tmy in the first dataset (MY) with the mi in the day to get the propotion of milk yield 'produced' on that day
    # all parts of the milking session before midnight are set to 0. The only milk yields in THIS dataset, are from the sessions completely produced in the current day AND the proportion produced after midnight.
    milking["mi_day_before"] = 0
    milking["tmy"] = milking["tmy"] * milking["mi_on_day"]
    milking["mylf"] = milking["mylf"] * milking["mi_on_day"]
    milking["myrf"] = milking["myrf"] * milking["mi_on_day"]
    milking["mylr"] = milking["mylr"] * milking["mi_on_day"]
    milking["myrr"] = milking["myrr"] * milking["mi_on_day"]

    # multiply the tmy in the second dataset (df_milking_add) with the mi on the day before to get the propotion of milk yield 'produced' on the previous day
    # all complete milk sessions and the part of the milking interval after midnight, are equaled to 0. The only milk yields in THIS dataset are milk yields from the proportion before midnight.
    MY_daily_add["mi_on_day"] = 0
    # change the DIM to the DIM of the previous day (so later you will add this MY to the corresponding day - before midnight)
    MY_daily_add["dim_floor"] -= 1
    # if proportion of MI on the previous day is lower than 0, set equal to 0 (no milk yield assigned to the previous day)
    MY_daily_add[MY_daily_add["dim_floor"] < 0] = 0
    MY_daily_add["tmy"] = MY_daily_add["tmy"] * MY_daily_add["mi_day_before"]
    MY_daily_add["mylf"] = MY_daily_add["mylf"] * MY_daily_add["mi_day_before"]
    MY_daily_add["myrf"] = MY_daily_add["myrf"] * MY_daily_add["mi_day_before"]
    MY_daily_add["mylr"] = MY_daily_add["mylr"] * MY_daily_add["mi_day_before"]
    MY_daily_add["myrr"] = MY_daily_add["myrr"] * MY_daily_add["mi_day_before"]

    # combine both tables and lose unnecessary information
    # df_milking contains the data of milking sessions of the current day (full MI on current day & proportion produced on current day of 'overnight' MI)
    # df_milking_add contains data of milking sessions of the previous day (proportion produced on previous day of 'overnight' MI)
    milking = pd.concat([milking, MY_daily_add])
    milking = milking[
        [
            "farm_id",
            "animal_id",
            "lactation_id",
            "parity",
            "mi",
            "dim_floor",
            "tmy",
            "mylf",
            "myrf",
            "mylr",
            "myrr",
            "mi_day_before",
            "mi_on_day",
        ]
    ]
    del MY_daily_add

    # multiply the mi with the proportion of each day to get the true values of mi per period
    # In df_milking (contains current day info): mi_day_before is always 0, mi_on_day is either 1 (full MI on current day) or smaller than 1 (proportion produced on current day of 'overnight' MI)
    # In df_milking_add (contains previous day): mi_on_day is always 0, mi_day_before lies between 0 and 1 (proportion produced on previous day)
    milking["mi"] = milking["mi"] * (milking["mi_day_before"] + milking["mi_on_day"])

    # group by dim_floor to get the daily milk yields. Add all measurements to the assigned day
    MY_daily = milking.groupby(
        ["farm_id", "animal_id", "lactation_id", "parity", "dim_floor"], dropna=False
    ).sum()
    MY_daily.reset_index(inplace=True)
    MY_daily = MY_daily.rename(
        columns={
            "dim_floor": "dim",
            "tmy": "tdmy",
            "mylf": "dmylf",
            "myrf": "dmyrf",
            "mylr": "dmylr",
            "myrr": "dmyrr",
        }
    )
    del milking

    # add 1 to dim to avoid errors during the fitting process (allow any y offset, might cause problems if y=0 in some models)
    MY_daily["dim"] += 1

    # correct the milk yields to true daily milk yield by deviding through the mi for each my calculation and multiply by 24h (correct for missing data)
    MY_daily["tdmy"] = (MY_daily["tdmy"] / MY_daily["mi"]) * 24
    MY_daily["dmylf"] = (MY_daily["dmylf"] / MY_daily["mi"]) * 24
    MY_daily["dmyrf"] = (MY_daily["dmyrf"] / MY_daily["mi"]) * 24
    MY_daily["dmylr"] = (MY_daily["dmylr"] / MY_daily["mi"]) * 24
    MY_daily["dmyrr"] = (MY_daily["dmyrr"] / MY_daily["mi"]) * 24

    MY_daily = MY_daily[
        [
            "farm_id",
            "animal_id",
            "lactation_id",
            "parity",
            "dim",
            "tdmy",
            "dmylf",
            "dmyrf",
            "dmylr",
            "dmyrr",
        ]
    ]

    # calculate a 7 day rolling median of the tdmy and select for tdmy values that are lower than
    # 1.5*the rolling median and below 75 kg daily milk yield
    MY_daily["tdmy7dm"] = MY_daily["tdmy"].rolling(7).median()
    MY_daily.loc[(MY_daily["dim"] < 7), "tdmy7dm"] = MY_daily.loc[
        (MY_daily["dim"] < 7), "tdmy"
    ]
    MY_daily = MY_daily[(MY_daily["tdmy"] < 1.5 * MY_daily["tdmy7dm"])]
    MY_daily = MY_daily[(MY_daily["tdmy"] < 75)]

    MY_daily = MY_daily[
        [
            "farm_id",
            "animal_id",
            "lactation_id",
            "parity",
            "dim",
            "tdmy",
            "dmylf",
            "dmyrf",
            "dmylr",
            "dmyrr",
        ]
    ]

    return MY_daily



import json
import os
import re
from pathlib import Path

import pandas as pd


def mergeTables_single(rootdir: str | Path, farmname: str, newtable: str) -> None:
    """
    Merge original tables; create tables with new database scheme

    Parameters
    ----------
    rootdir : str or Path
        Filepath to CowBase root directory (e.g. linux: '/var/lib/'; windows: 'C:\\Users\\user\\Documents\\')
    farmname : str
        name of farm to analyze (e.g., 'd_farmname')
    newtable : str
        name of the table to merge
    """

    rootdir = Path(rootdir)
    farmtype = farmname.split("_")[0]

    # Read in all necessary setting files and dictionaries
    with open(rootdir / "config" / "M3_tableDict.json") as file:
        config_json = json.load(file)
        infodict1 = config_json[farmtype]
        infodict2 = config_json["d_37"]
    if newtable not in list(infodict1.keys()):
        print(f"Table {newtable} is not available for farming system {farmtype}!")
        return

    # Initiate savepath
    backupdates = []
    df_data = pd.DataFrame({})
    filepath = rootdir / "OUTPUT" / farmname / "01_full_table"
    savepath = rootdir / "OUTPUT" / farmname / "02_merged_table"
    savepath.mkdir(parents=True, exist_ok=True)

    # Iterate over all backups, read data for each folder (with correct version dictionary) and merge temporal
    for folders in os.listdir(filepath):
        infodict = infodict2 if folders[-3:] == "_37" else infodict1
        df_data_add = pd.DataFrame({})
        df_data_add, savefilename_new, backupdates = readAndMerge(
            infodict, newtable, backupdates, filepath, folders
        )

        if savefilename_new == None:
            continue
        else:
            savefilename = savefilename_new
            df_data_add = df_data_add[infodict[newtable]["rename"].keys()]
            df_data_add.rename(columns=infodict[newtable]["rename"], inplace=True)
            df_data = pd.concat([df_data, df_data_add], ignore_index=True)
            df_data = df_data.drop_duplicates(
                subset=infodict[newtable]["drop_duplicates"], keep="last"
            )
    if df_data.empty == False:
        if newtable == "milking" and (farmtype == "d" or farmtype == "d_37"):
            df_data = calculatingMilkDeLaval(df_data)
        elif newtable == "milking" and farmtype == "l":
            df_data = calculatingMilkLely(df_data)
        df_data.to_csv(savepath / savefilename, sep=";", index=False)


def readAndMerge(
    infodict: dict,
    newtable: str,
    backupdates: list,
    filepath: str | Path,
    folders: str | Path,
) -> tuple[pd.DataFrame, Path | None, list]:
    """
    For a single table, iterate over all files in the folder. If there is data available, read and return

    Parameters
    ----------
    infodict : dict
        Dictionary containing single original tables and columns to open and merge
    newtable : str
        name of the table to merge
    backupdates : list
        list of all backupdates
    filepath : str or Path
        filepath to the dictionary containing backups of a single farm
    folders : str or Path
        name of the folder to run in

    Return
    ------
    df_data_add : dataframe
        dataframe containing data of a single new database table of a single backup
    savefilename : Path
        name to save complete files
    backupdates : list
        list of all backupdates
    """

    # Iterate over all separate raw tables that are needed to create a new database table
    for tables in list(infodict[newtable]["data"].keys()):
        table_available = False
        for files in os.listdir(filepath / folders):
            tablename = re.split("_", files[:-4])[-1]
            if tablename == tables:
                table_available = True
                splitstr = re.split("_", files)
                backupdate = int(re.split("_", files[:-4])[2])
                backupdates.append(backupdate)
                savefilename = f"{splitstr[0]}_{splitstr[1]}_{min(backupdates)}_{max(backupdates)}_{newtable}.csv"

                # load data for table and select for certain columns given in the tabledict
                df_data_file = pd.read_csv(
                    filepath / folders / files,
                    delimiter=";",
                    header=0,
                    engine="c",
                )
        if table_available == False:
            print(f"Skipped {newtable}! Data is not available!")
            return pd.DataFrame({}), None, backupdates
        if "on" in list(infodict[newtable]["data"][tables].keys()):
            if "prerename" in list(infodict[newtable]["data"][tables].keys()):
                df_data_file.rename(
                    columns=infodict[newtable]["data"][tables]["prerename"],
                    inplace=True,
                )
            df_data_add = df_data_add.merge(
                df_data_file,
                on=infodict[newtable]["data"][tables]["on"],
                how=infodict[newtable]["data"][tables]["how"],
            )

        else:
            df_data_add = df_data_file.copy()

    return df_data_add, savefilename, backupdates


def calculatingMilkDeLaval(df_data: pd.DataFrame) -> pd.DataFrame:
    """
    For DeLaval, the interval between previous milkings and current milkings are calculated

    Parameters
    ----------
    df_data : dataframe
        dataframe of milking data without milking interval

    Return
    ------
    df_data : dataframe
        dataframe of milking data with milking interval
    """

    df_data[["ended_at", "previously_ended_at"]] = df_data[
        ["ended_at", "previously_ended_at"]
    ].apply(pd.to_datetime)
    df_data.loc[
        ~pd.isnull(df_data.ended_at) & ~pd.isnull(df_data.previously_ended_at),
        "mi",
    ] = (
        (
            df_data.loc[
                ~pd.isnull(df_data.ended_at) & ~pd.isnull(df_data.previously_ended_at),
                "ended_at",
            ]
            - df_data.loc[
                ~pd.isnull(df_data.ended_at) & ~pd.isnull(df_data.previously_ended_at),
                "previously_ended_at",
            ]
        ).dt.total_seconds()
        / 3600
    ).round(
        decimals=2
    )
    return df_data


def calculatingMilkLely(df_data: pd.DataFrame) -> pd.DataFrame:
    """
    For Lely, the time of the previous milkings are calculated based on the milking interval and the current times
    Further the Electrical Conductivity is transformed from Lely units to S/m

    Parameters
    ----------
    df_data : dataframe
        dataframe of milking data without previous milkting time and EC

    Return
    ------
    df_data : dataframe
        dataframe of milking data with previous milkting time and EC
    """

    df_data[["ended_at"]] = df_data[["ended_at"]].apply(pd.to_datetime)
    df_data.mi = pd.to_timedelta(df_data.mi.astype(float), unit="s")
    df_data.loc[
        ~pd.isnull(df_data.ended_at) & ~pd.isnull(df_data.mi),
        "previously_ended_at",
    ] = (
        df_data.loc[
            ~pd.isnull(df_data.ended_at) & ~pd.isnull(df_data.mi),
            "ended_at",
        ]
        - df_data.loc[
            ~pd.isnull(df_data.ended_at) & ~pd.isnull(df_data.mi),
            "mi",
        ]
    )
    df_data["mi"] = ((df_data["mi"] / pd.to_timedelta(1, unit="D")) * 24).round(2)

    df_data["eclf"] = ((0.084 * df_data.eclf_lely) - 1).round(2)
    df_data["ecrf"] = ((0.084 * df_data.ecrf_lely) - 1).round(2)
    df_data["eclr"] = ((0.084 * df_data.eclr_lely) - 1).round(2)
    df_data["ecrr"] = ((0.084 * df_data.ecrr_lely) - 1).round(2)

    df_data.loc[df_data.eclf <= 0, "eclf"] = 0
    df_data.loc[df_data.ecrf <= 0, "ecrf"] = 0
    df_data.loc[df_data.eclr <= 0, "eclr"] = 0
    df_data.loc[df_data.ecrr <= 0, "ecrr"] = 0

    return df_data


def mergeTables_all(rootdir: str | Path, farmname: str) -> None:
    """
    Merge original tables; create tables with new database scheme

    Parameters
    ----------
    rootpath : str or Path
        Filepath to CowBase root directory (e.g. linux: '/var/lib/'; windows: 'C:\\Users\\user\\Documents\\')
    farmname : str
        name of farm to analyze (e.g., 'd_farmname')

    """

    rootdir = Path(rootdir)

    farmtype = farmname.split("_")[0]

    if farmtype == "d":
        tablelist = [
            "animal",
            "lactation",
            "milking",
            "herdnavigator",
            "insemination",
            "bcs",
            "cleaning",
        ]
    elif farmtype == "l":
        tablelist = ["animal", "lactation", "milking", "activity"]

    for tables in tablelist:
        mergeTables_single(rootdir, farmname, tables)
